from typing import List

from sampo.scheduler.base import Scheduler, SchedulerType
from sampo.scheduler.resource.average_req import AverageReqResourceOptimizer
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor
from sampo.schemas.exceptions import NoSufficientContractorError
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time


class AverageBinarySearchResourceOptimizingScheduler:

    def __init__(self, base_scheduler: Scheduler):
        self._base_scheduler = base_scheduler
        self._resource_optimizer = AverageReqResourceOptimizer()
        base_scheduler.resource_optimizer = self._resource_optimizer

    def schedule_with_cache(self, wg: WorkGraph, contractors: List[Contractor], deadline: Time,
                            spec: ScheduleSpec = ScheduleSpec(),
                            validate: bool = False, assigned_parent_time: Time = Time(0)) \
            -> tuple[Schedule, Time, Timeline, list[GraphNode]]:
        def call_scheduler(k) -> tuple[Schedule, Time, Timeline, list[GraphNode]]:
            self._resource_optimizer.k = k
            try:
                return self._base_scheduler.schedule_with_cache(wg, contractors, spec, validate, assigned_parent_time)
            except NoSufficientContractorError:
                return None, Time.inf(), None, None

        def fitness(k):
            result = call_scheduler(k)[1]
            if result > deadline:
                result = Time.inf()
            return result

        k_min = 1
        k_max = 10000

        while k_max - k_min > 0.05:
            m = (k_min + k_max) / 2
            time_m = fitness(m)
            if time_m.is_inf():
                k_max = m
            else:
                k_min = m

        return call_scheduler(k_min)



from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

from sampo.schemas.contractor import Contractor
from sampo.schemas.resources import Equipment, ConstructionObject, Worker
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.schemas.works import WorkUnit
from sampo.utilities.serializers import custom_serializer


# TODO: describe the class (description, parameters)
@dataclass
class ScheduledWork(AutoJSONSerializable['ScheduledWork']):

    ignored_fields = ['equipments', 'materials', 'object']

    def __init__(self,
                 work_unit: WorkUnit,
                 start_end_time: Tuple[Time, Time],
                 workers: List[Worker],
                 contractor: Contractor | str,
                 equipments: Optional[List[Equipment]] = None,
                 materials: Optional[List[Equipment]] = None,
                 object: Optional[ConstructionObject] = None):
        self.work_unit = work_unit
        self.start_end_time = start_end_time
        self.workers = workers
        self.equipments = equipments
        self.materials = materials
        self.object = object
        if contractor is not None:
            if isinstance(contractor, str):
                self.contractor = contractor
            else:
                self.contractor = contractor.name if contractor.name else contractor.id
        else:
            self.contractor = ""

        self.cost = 0
        for worker in self.workers:
            self.cost += worker.get_cost() * self.duration.value

    def __str__(self):
        return f'ScheduledWork[work_unit={self.work_unit}, start_end_time={self.start_end_time}, ' \
               f'workers={self.workers}, contractor={self.contractor}]'

    def __repr__(self):
        return self.__str__()

    @custom_serializer('workers')
    @custom_serializer('start_end_time')
    def serialize_serializable_list(self, value):
        return [t._serialize() for t in value]

    @classmethod
    @custom_serializer('start_end_time', deserializer=True)
    def deserialize_time(cls, value):
        return [Time._deserialize(t) for t in value]

    @classmethod
    @custom_serializer('workers', deserializer=True)
    def deserialize_workers(cls, value):
        return [Worker._deserialize(t) for t in value]

    # TODO: describe the function (description, parameters, return type)
    def get_actual_duration(self, work_estimator: Optional[WorkTimeEstimator] = None) -> Time:
        return self.work_unit.estimate_static(self.workers, work_estimator)

    # TODO: describe the function (description return type)
    @property
    def start_time(self) -> Time:
        return self.start_end_time[0]

    # TODO: describe the function (description, parameters, return type)
    @start_time.setter
    def start_time(self, val: Time):
        self.start_end_time = (val, self.start_end_time[1])

    # TODO: describe the function (description, return type)
    @property
    def finish_time(self) -> Time:
        return self.start_end_time[1]

    @property
    def min_child_start_time(self) -> Time:
        return self.finish_time if self.work_unit.is_service_unit else self.finish_time + 1

    # TODO: describe the function (description, parameters, return type)
    @finish_time.setter
    def finish_time(self, val: Time):
        self.start_end_time = (self.start_end_time[0], val)

    # TODO: describe the function (description, return type)
    @staticmethod
    def start_time_getter():
        return lambda x: x.start_end_time[0]

    # TODO: describe the function (description, return type)
    @staticmethod
    def finish_time_getter():
        return lambda x: x.start_end_time[1]

    # TODO: describe the function (description, return type)
    @property
    def duration(self) -> Time:
        start, end = self.start_end_time
        return end - start
    
    # TODO: describe the function (description, parameters, return type)
    def is_overlapped(self, time: int) -> bool:
        start, end = self.start_end_time
        return start <= time < end

    # TODO: describe the function (description, return type)
    def to_dict(self) -> Dict[str, Any]:
        return {
            "task_id": self.work_unit.id,
            "task_name": self.work_unit.name,
            "start": self.start_time.value,
            "finish": self.finish_time.value,
            "contractor_id": self.contractor,
            "workers": {worker.name: worker.count for worker in self.workers},
        }

    def __deepcopy__(self, memodict={}):
        return ScheduledWork(deepcopy(self.work_unit, memodict),
                             deepcopy(self.start_end_time, memodict),
                             deepcopy(self.workers, memodict),
                             self.contractor)



import random
from typing import Optional, List, Tuple, Hashable

from sampo.schemas.graph import GraphNode, EdgeType
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.works import WorkUnit
from sampo.utilities.sampler.requirements import get_worker_reqs_list
from sampo.utilities.sampler.types import MinMax
from sampo.utilities.sampler.works import get_work_unit, get_similar_work_unit


class Sampler:
    def __init__(self,
                 seed: Optional[Hashable] = None
                 ):
        self.rand = random.Random(seed)

    def worker_reqs(self,
                    volume: Optional[MinMax[int]] = MinMax[int](1, 50),
                    worker_count: Optional[MinMax[int]] = MinMax[int](1, 100)
                    ) -> List[WorkerReq]:
        return get_worker_reqs_list(self.rand, volume, worker_count)

    def work_unit(self,
                  name: str,
                  work_id: Optional[str] = '',
                  volume_type: Optional[str] = 'unit',
                  group: Optional[str] = "default",
                  work_volume: Optional[MinMax[float]] = MinMax[float](0.1, 100.0),
                  req_volume: Optional[MinMax[int]] = MinMax[int](1, 50),
                  req_worker_count: Optional[MinMax[int]] = MinMax[int](1, 100)
                  ) -> WorkUnit:
        return get_work_unit(self.rand, name, work_id, volume_type, group, work_volume, req_volume, req_worker_count)

    def similar_work_unit(self,
                          exemplar: WorkUnit,
                          scalar: Optional[float] = 1.0,
                          name: Optional[str] = '',
                          work_id: Optional[str] = ''
                          ) -> WorkUnit:
        return get_similar_work_unit(self.rand, exemplar, scalar, name, work_id)

    def graph_node(self,
                   name: str,
                   edges: List[Tuple[GraphNode, float, EdgeType]],
                   work_id: Optional[str] = '',
                   volume_type: Optional[str] = 'unit',
                   group: Optional[str] = "default",
                   work_volume: Optional[MinMax[float]] = MinMax[float](0.1, 100.0),
                   req_volume: Optional[MinMax[int]] = MinMax[int](1, 50),
                   req_worker_count: Optional[MinMax[int]] = MinMax[int](1, 100)
                   ) -> GraphNode:
        wu = get_work_unit(self.rand, name, work_id, volume_type, group, work_volume, req_volume, req_worker_count)
        return GraphNode(wu, edges)

    def similar_graph_node(self,
                           exemplar: GraphNode,
                           edges: List[Tuple[GraphNode, float, EdgeType]],
                           scalar: Optional[float] = 1.0,
                           name: Optional[str] = '',
                           work_id: Optional[str] = ''
                           ) -> GraphNode:
        wu = get_similar_work_unit(self.rand, exemplar.work_unit, scalar, name, work_id)
        return GraphNode(wu, edges)



import math
import random
from typing import Dict, List, Tuple, Optional, Callable

from deap.base import Toolbox

from sampo.scheduler.base import Scheduler, SchedulerType
from sampo.scheduler.genetic.operators import FitnessFunction, TimeFitness
from sampo.scheduler.genetic.schedule_builder import build_schedule
from sampo.scheduler.heft.base import HEFTScheduler, HEFTBetweenScheduler
from sampo.scheduler.heft.prioritization import prioritization
from sampo.scheduler.resource.average_req import AverageReqResourceOptimizer
from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.resource.identity import IdentityResourceOptimizer
from sampo.scheduler.resources_in_time.average_binary_search import AverageBinarySearchResourceOptimizingScheduler
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor, get_worker_contractor_pool
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.validation import validate_schedule


class GeneticScheduler(Scheduler):

    def __init__(self,
                 number_of_generation: Optional[int] = 50,
                 size_selection: Optional[int or None] = None,
                 mutate_order: Optional[float or None] = None,
                 mutate_resources: Optional[float or None] = None,
                 size_of_population: Optional[float or None] = None,
                 rand: Optional[random.Random] = None,
                 seed: Optional[float or None] = None,
                 n_cpu: int = 1,
                 fitness_constructor: Callable[[Toolbox], FitnessFunction] = TimeFitness,
                 scheduler_type: SchedulerType = SchedulerType.Genetic,
                 resource_optimizer: ResourceOptimizer = IdentityResourceOptimizer(),
                 work_estimator: Optional[WorkTimeEstimator or None] = None):
        super().__init__(scheduler_type=scheduler_type,
                         resource_optimizer=resource_optimizer,
                         work_estimator=work_estimator)
        self.number_of_generation = number_of_generation
        self.size_selection = size_selection
        self.mutate_order = mutate_order
        self.mutate_resources = mutate_resources
        self.size_of_population = size_of_population
        self.rand = rand or random.Random(seed)
        self.fitness_constructor = fitness_constructor
        self.work_estimator = work_estimator
        self._n_cpu = n_cpu

        self._time_border = None
        self._deadline = None

    def __str__(self) -> str:
        return f'GeneticScheduler[' \
               f'generations={self.number_of_generation},' \
               f'size_selection={self.size_selection},' \
               f'mutate_order={self.mutate_order},' \
               f'mutate_resources={self.mutate_resources}' \
               f']'

    def get_params(self, works_count: int) -> Tuple[int, float, float, int]:
        size_selection = self.size_selection
        if size_selection is None:
            if works_count < 300:
                size_selection = 20
            else:
                size_selection = works_count // 15

        mutate_order = self.mutate_order
        if mutate_order is None:
            if works_count < 300:
                mutate_order = 0.05
            else:
                mutate_order = 2 / math.sqrt(works_count)

        mutate_resources = self.mutate_resources
        if mutate_resources is None:
            if works_count < 300:
                mutate_resources = 0.1
            else:
                mutate_resources = 6 / math.sqrt(works_count)

        size_of_population = self.size_of_population
        if size_of_population is None:
            if works_count < 300:
                size_of_population = 20
            elif 1500 > works_count >= 300:
                size_of_population = 50
            else:
                size_of_population = works_count // 50
        return size_selection, mutate_order, mutate_resources, size_of_population

    def set_use_multiprocessing(self, n_cpu: int):
        self._n_cpu = n_cpu

    def set_time_border(self, time_border: int):
        self._time_border = time_border

    def set_deadline(self, deadline: Time):
        self._deadline = deadline

    def generate_first_population(self, wg: WorkGraph, contractors: list[Contractor]):
        if self._deadline is None:
            def init_schedule(scheduler_class):
                return (scheduler_class(work_estimator=self.work_estimator).schedule(wg, contractors),
                        list(reversed(prioritization(wg, self.work_estimator))))

            def init_k_schedule(scheduler_class, k):
                return (scheduler_class(work_estimator=self.work_estimator,
                                        resource_optimizer=AverageReqResourceOptimizer(k)).schedule(wg, contractors),
                        list(reversed(prioritization(wg, self.work_estimator))))

            return {
                "heft_end": init_schedule(HEFTScheduler),
                "heft_between": init_schedule(HEFTBetweenScheduler),
                "12.5%": init_k_schedule(HEFTScheduler, 8),
                "25%": init_k_schedule(HEFTScheduler, 4),
                "75%": init_k_schedule(HEFTScheduler, 4 / 3),
                "87.5%": init_k_schedule(HEFTScheduler, 8 / 7)
            }
        else:
            def init_schedule(scheduler_class):
                schedule = AverageBinarySearchResourceOptimizingScheduler(
                    scheduler_class(work_estimator=self.work_estimator)
                ).schedule_with_cache(wg, contractors, self._deadline)[0]
                return schedule, list(reversed(prioritization(wg, self.work_estimator)))

            return {
                "heft_end": init_schedule(HEFTScheduler),
                "heft_between": init_schedule(HEFTBetweenScheduler)
            }

    def schedule_with_cache(self, wg: WorkGraph,
                            contractors: List[Contractor],
                            spec: ScheduleSpec = ScheduleSpec(),
                            validate: bool = False,
                            assigned_parent_time: Time = Time(0),
                            timeline: Timeline | None = None) \
            -> tuple[Schedule, Time, Timeline, list[GraphNode]]:

        init_schedules = self.generate_first_population(wg, contractors)

        size_selection, mutate_order, mutate_resources, size_of_population = self.get_params(wg.vertex_count)
        worker_pool = get_worker_contractor_pool(contractors)

        scheduled_works, schedule_start_time, timeline, order_nodes = build_schedule(wg,
                                                                                     contractors,
                                                                                     worker_pool,
                                                                                     size_of_population,
                                                                                     self.number_of_generation,
                                                                                     size_selection,
                                                                                     mutate_order,
                                                                                     mutate_resources,
                                                                                     init_schedules,
                                                                                     self.rand,
                                                                                     spec,
                                                                                     self.fitness_constructor,
                                                                                     self.work_estimator,
                                                                                     n_cpu=self._n_cpu,
                                                                                     assigned_parent_time=assigned_parent_time,
                                                                                     timeline=timeline,
                                                                                     time_border=self._time_border)
        schedule = Schedule.from_scheduled_works(scheduled_works.values(), wg)

        if validate:
            validate_schedule(schedule, wg, contractors)

        return schedule, schedule_start_time, timeline, order_nodes



import multiprocessing as mp
from typing import Callable, Optional
from uuid import uuid4

import numpy as np

from sampo.schemas.contractor import WorkerContractorPool
from sampo.schemas.interval import IntervalGaussian
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.schemas.works import WorkUnit


def init_worker(wte_constructor, wte_args):
    global wte
    wte = wte_constructor(*wte_args)


def calculate_time(args: tuple[list[int], list[str], list[IntervalGaussian | None]]) -> int:
    counts, names, productivities = args
    work_unit = WorkUnit("", "")
    contractor = str(uuid4())
    worker_team = [Worker(str(uuid4()), name, count, contractor, productivity)
                   for name, count, productivity in zip(names, counts, productivities)]
    return work_unit.estimate_static(worker_team, wte).value


# gets the 'next' team
def inc(team: list[int], max_counts: list[int]):
    i = len(team) - 1
    while i >= 0:
        team[i] += 1
        if team[i] > max_counts[i]:
            team[i] = 0
            i -= 1
        else:
            break


def calculate_arg_results(worker_labels: list[str],
                          worker_counts: list[int],
                          work_time_estimator: WorkTimeEstimator):
    n_workers = len(worker_labels)
    iterations = 1  # teams count
    for v in worker_counts:
        if v != 0:
            iterations *= v + 1

    count_args = []
    cur_team = [0 for _ in range(n_workers)]
    productivities = [None for _ in range(n_workers)]

    for i in range(iterations):
        count_args.append((cur_team[:], worker_labels, productivities))
        inc(cur_team, worker_counts)

    with mp.Pool(initializer=init_worker,
                 initargs=work_time_estimator.split_to_constructor_and_params()) as p:
        results = p.map(calculate_time, count_args)

        with open('trained.txt', 'w') as f:
            for res, count_arg in zip(results, count_args):
                cur_team = count_arg[0]
                f.write(' '.join(map(str, cur_team)))
                f.write(f' {res}')
                f.write('\n')


class TestWorkTimeEstimator(WorkTimeEstimator):

    def set_mode(self, use_idle: Optional[bool] = True, mode: Optional[str] = 'realistic'):
        pass

    def estimate_time(self, work_name: str, work_volume: float, resources: WorkerContractorPool) -> Time:
        return Time(int(work_volume))

    def split_to_constructor_and_params(self) -> tuple[Callable[[...], 'WorkTimeEstimator'], tuple]:
        return TestWorkTimeEstimator, ()


def get(teams: np.ndarray, team: list[int]) -> int:
    cur = teams

    for next_worker in team:
        cur = cur[next_worker]

    return cur


def train_surrogate(teams: np.ndarray):
    a1 = a2 = a3 = a4 = 1




if __name__ == '__main__':
    worker_labels = ['1', '2', '3', '4', '5']
    worker_counts = [ 1 ,  1 ,  1 ,  1 ,  1 ]

    calculate_arg_results(worker_labels, worker_counts, TestWorkTimeEstimator())
